use crate::app_config::AppConfig;
use crate::crypto::pki;
use crate::devices::device::{DeviceId, DeviceInfo};
use crate::energy::energy_actor;
use crate::server::custom_error::HttpResult;
use crate::server::WebEnergyActor;
use actix_web::{web, HttpResponse};
use jsonwebtoken::{Algorithm, DecodingKey, Validation};
use openssl::nid::Nid;
use openssl::x509::{X509Req, X509};
use std::collections::HashSet;

#[derive(Debug, serde::Deserialize)]
pub struct EnrollRequest {
    /// Device CSR
    csr: String,
    /// Associated device information
    info: DeviceInfo,
}

/// Enroll a new device
pub async fn enroll(req: web::Json<EnrollRequest>, actor: WebEnergyActor) -> HttpResult {
    // Check device information
    if let Some(e) = req.info.error() {
        log::error!("Failed to validate device information! {e}");
        return Ok(HttpResponse::BadRequest().json(e));
    }

    // Check CSR
    let csr = match X509Req::from_pem(req.csr.as_bytes()) {
        Ok(r) => r,
        Err(e) => {
            log::error!("Failed to parse given CSR! {e}");
            return Ok(HttpResponse::BadRequest().json("Failed to parse given CSR!"));
        }
    };

    if !csr.verify(csr.public_key()?.as_ref())? {
        log::error!("Invalid CSR signature!");
        return Ok(HttpResponse::BadRequest().json("Could not verify CSR signature!"));
    }

    let cn = match csr.subject_name().entries_by_nid(Nid::COMMONNAME).next() {
        None => {
            log::error!("Missing Common Name in CSR!");
            return Ok(HttpResponse::BadRequest().json("Missing Common Name in CSR!"));
        }
        Some(cn) => cn.data().as_utf8()?.to_string(),
    };

    if !lazy_regex::regex!("[a-zA-Z0-9 ]{1,100}").is_match(&cn) {
        log::error!("Given Common Name is invalid!");
        return Ok(HttpResponse::BadRequest().json("Invalid Common Name in CSR!"));
    }

    let device_id = DeviceId(cn);
    log::info!(
        "Received enrollment request for device with ID {device_id:?} - {:#?}",
        req.info
    );

    if actor
        .send(energy_actor::CheckDeviceExists(device_id.clone()))
        .await?
    {
        log::error!("Device could not be enrolled: it already exists!");
        return Ok(
            HttpResponse::Conflict().json("A device with the same ID has already been enrolled!")
        );
    }

    actor
        .send(energy_actor::EnrollDevice(device_id, req.0.info, csr))
        .await??;

    Ok(HttpResponse::Accepted().json("Device successfully enrolled"))
}

#[derive(serde::Deserialize)]
pub struct ReqWithDevID {
    id: DeviceId,
}

#[derive(serde::Serialize)]
#[serde(tag = "status")]
enum EnrollmentDeviceStatus {
    Unknown,
    Pending,
    Validated,
}

/// Check device enrollment status
pub async fn enrollment_status(
    query: web::Query<ReqWithDevID>,
    actor: WebEnergyActor,
) -> HttpResult {
    let dev = actor
        .send(energy_actor::GetSingleDevice(query.id.clone()))
        .await?;

    let status = match dev {
        None => EnrollmentDeviceStatus::Unknown,
        Some(d) if d.validated => EnrollmentDeviceStatus::Validated,
        _ => EnrollmentDeviceStatus::Pending,
    };

    Ok(HttpResponse::Ok().json(status))
}

/// Get device certificate
pub async fn get_certificate(query: web::Query<ReqWithDevID>, actor: WebEnergyActor) -> HttpResult {
    let dev = actor
        .send(energy_actor::GetSingleDevice(query.id.clone()))
        .await?;

    let dev = match dev {
        Some(d) if d.validated => d,
        _ => {
            log::error!("Device attempted to retrieve an unavailable certificate!");
            return Ok(HttpResponse::UnprocessableEntity().json("Certificate not available yet!"));
        }
    };

    let cert = std::fs::read(AppConfig::get().device_cert_path(&dev.id))?;

    Ok(HttpResponse::Ok()
        .content_type("application/x-pem-file")
        .body(cert))
}

#[derive(serde::Deserialize)]
pub struct SyncRequest {
    payload: String,
}

#[derive(Debug, serde::Serialize, serde::Deserialize)]
struct Claims {
    info: DeviceInfo,
}

/// Synchronize device
pub async fn sync_device(body: web::Json<SyncRequest>, actor: WebEnergyActor) -> HttpResult {
    // First, we need to extract device kid from query
    let Ok(jwt_header) = jsonwebtoken::decode_header(&body.payload) else {
        log::error!("Failed to decode JWT header!");
        return Ok(HttpResponse::BadRequest().json("Failed to decode JWT header!"));
    };

    let Some(kid) = jwt_header.kid else {
        log::error!("Missing KID in JWT!");
        return Ok(HttpResponse::BadRequest().json("Missing KID in JWT!"));
    };

    // Fetch device information
    let Some(device) = actor
        .send(energy_actor::GetSingleDevice(DeviceId(kid)))
        .await?
    else {
        log::error!("Sent a JWT for a device which does not exists!");
        return Ok(HttpResponse::NotFound().json("Sent a JWT for a device which does not exists!"));
    };

    if !device.validated {
        log::error!("Sent a JWT for a device which is not validated!");
        return Ok(HttpResponse::PreconditionFailed()
            .json("Sent a JWT for a device which is not validated!"));
    }

    // Check certificate revocation status
    let cert_bytes = std::fs::read(AppConfig::get().device_cert_path(&device.id))?;
    let certificate = X509::from_pem(&cert_bytes)?;

    if pki::CertData::load_devices_ca()?.is_revoked(&certificate)? {
        log::error!("Sent a JWT using a revoked certificate!");
        return Ok(
            HttpResponse::PreconditionFailed().json("Sent a JWT using a revoked certificate!")
        );
    }

    let (key, alg) = match DecodingKey::from_ec_pem(&cert_bytes) {
        Ok(key) => (key, Algorithm::ES256),
        Err(e) => {
            log::warn!("Failed to decode certificate as EC certificate {e}, trying RSA...");
            (
                DecodingKey::from_rsa_pem(&cert_bytes).expect("Failed to decode RSA certificate"),
                Algorithm::RS256,
            )
        }
    };
    let mut validation = Validation::new(alg);
    validation.validate_exp = false;
    validation.required_spec_claims = HashSet::default();

    let c = match jsonwebtoken::decode::<Claims>(&body.payload, &key, &validation) {
        Ok(c) => c,
        Err(e) => {
            log::error!("Failed to validate JWT! {e}");
            return Ok(HttpResponse::PreconditionFailed().json("Failed to validate JWT!"));
        }
    };

    let res = actor
        .send(energy_actor::SynchronizeDevice(device.id, c.claims.info))
        .await??;

    Ok(HttpResponse::Ok().json(res))
}
